/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with {library}; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package ch.hortis.sonar.mvn;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.sql.Connection;
import java.util.Date;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.FlushModeType;
import javax.persistence.NoResultException;

import org.apache.maven.model.CiManagement;
import org.apache.maven.model.IssueManagement;
import org.apache.maven.model.Model;
import org.apache.maven.model.Scm;
import org.apache.maven.model.io.xpp3.MavenXpp3Reader;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.codehaus.plexus.util.xml.pull.XmlPullParserException;

import ch.hortis.sonar.jpa.Persistence;
import ch.hortis.sonar.jpa.PersistenceException;
import ch.hortis.sonar.model.Collectable;
import ch.hortis.sonar.model.JdbcData;
import ch.hortis.sonar.model.MavenProject;
import ch.hortis.sonar.model.MetricMeasure;
import ch.hortis.sonar.model.ProjectLink;
import ch.hortis.sonar.model.Snapshot;
import ch.hortis.sonar.model.SnapshotGroup;
import ch.hortis.sonar.model.WithFile;
import ch.hortis.sonar.mvn.mc.FilesRepository;
import ch.hortis.sonar.mvn.mc.FilesRepositoryImpl;
import ch.hortis.sonar.mvn.mc.MeasuresCollector;
import ch.hortis.sonar.service.MavenProjectService;
import ch.hortis.sonar.service.MetricService;
import ch.hortis.sonar.service.RulesService;
import ch.hortis.sonar.service.SnapshotGroupService;
import ch.hortis.sonar.service.WebInterfaceService;
import ch.hortis.sonar.service.WebInterfaceServiceImpl;

/**
 * Main sonar plugin reports for plugins metrics parsing and historic generation
 *
 * @goal collect
 */
public class CollectMojo extends AbstractMojo {

  public static boolean firstCall = false;

  /**
   * The maven project running this plugin
   *
   * @parameter expression="${project}"
   * @required
   * @readonly
   */
  private org.apache.maven.project.MavenProject mavenProject;

  /**
   * JDBC URL
   *
   * @parameter expression="${sonar.jdbc.url}" default-value="jdbc:derby://localhost:1527/sonar" alias="sonar.jdbc.url"
   */
  private String jdbcURL;

  /**
   * JDBC driver class
   *
   * @parameter expression="${sonar.jdbc.driver}" default-value="org.apache.derby.jdbc.ClientDriver" alias="sonar.jdbc.driver"
   */
  private String jdbcDriverClassName;

  /**
   * JDBC user name
   *
   * @parameter expression="${sonar.jdbc.username}" default-value="sonar" alias="sonar.jdbc.username"
   */
  private String jdbcUserName;

  /**
   * JDBC password
   *
   * @parameter expression="${sonar.jdbc.password}" default-value="sonar" alias="sonar.jdbc.password"
   */
  private String jdbcPassword;

  /**
   * Project branch
   *
   * @parameter expression="${branch}" default-value="HEAD" alias="branch"
   */
  private String branch;

  /**
   * Sonar host URL
   *
   * @parameter expression="${sonar.host.url}" default-value="http://localhost:9000" alias="sonar.host.url"
   */
  private String sonarHostURL;

  private EntityManager manager;

  private WebInterfaceService webInterfaceService;

  private FilesRepository filesRepository;

  private Date runDate;
  
  private List<Report> targetReports;
  
  public CollectMojo(){
    this.targetReports = Report.getReports();
  }
  
  public CollectMojo(List<Report> targetReports){
    this.targetReports = targetReports;
  }  

  public void setJdbcURL(String jdbcURL) {
    this.jdbcURL = jdbcURL;
  }

  public void setJdbcDriverClassName(String jdbcDriverClassName) {
    this.jdbcDriverClassName = jdbcDriverClassName;
  }

  public void setJdbcUserName(String jdbcUserName) {
    this.jdbcUserName = jdbcUserName;
  }

  public void setJdbcPassword(String jdbcPassword) {
    this.jdbcPassword = jdbcPassword;
  }

  public String getBranch() {
    return branch;
  }

  public void setBranch(String branch) {
    this.branch = branch;
  }

  public void setProject(org.apache.maven.project.MavenProject mavenProject) {
    this.mavenProject = mavenProject;
  }

  protected void setManager(EntityManager manager) {
    this.manager = manager;
  }

  public void setWebInterfaceService(WebInterfaceService webInterfaceService) {
    this.webInterfaceService = webInterfaceService;
  }

  public void setFilesRepository(FilesRepository filesRepository) {
    this.filesRepository = filesRepository;
  }
  
  protected List<Report> getTargetReports() {
    return targetReports;
  }

  public JdbcData getJdbcData() {
    JdbcData jdbcData = new JdbcData(jdbcURL, jdbcDriverClassName, jdbcUserName, jdbcPassword);
    jdbcData.setIsolationLevel(Connection.TRANSACTION_READ_COMMITTED);
    return jdbcData;
  }

  public Integer getCurrentSnapshotGroupId() {
    try {
      String groupId = System.getProperty(CollectMojo.class.getName() + ".snaphsotGroupId");
      if (groupId != null) {
        return new Integer(groupId);
      }
    } catch (SecurityException ex) {
      throw new IllegalStateException("Unable to retreive current snapshotgroup id due to security constraints");
    }
    return null;
  }

  private void setCurrentSnapshotGroupId(Integer id) {
    try {
      System.setProperty(CollectMojo.class.getName() + ".snaphsotGroupId", id.toString());
    } catch (SecurityException ex) {
      throw new IllegalStateException("Unable to set the current snapshotgroup id due to security constraints");
    }
  }

  public void execute() throws MojoExecutionException {
    boolean externalManager = (manager != null);
    runDate = new Date();
    if (!externalManager) {
      manager = getPersistence().getNewEntityManager();
    }
    if (webInterfaceService == null) {
      webInterfaceService = new WebInterfaceServiceImpl(sonarHostURL);
    }
    MavenProject sonarProject;
    try {
      manager.getTransaction().begin();
      sonarProject = saveMavenProject(mavenProject, null);
      manager.getTransaction().commit();
    } catch (Exception ex) {
      manager.getTransaction().rollback();
      throw new MojoExecutionException("Error occured during maven project DB persistance", ex);
    }

    if (firstCall && sonarProject.getParent() != null) {
      MavenProject top = sonarProject.getParent();
      while (top.getParent() != null) {
        top = top.getParent();
      }
      getLog().info("Sonar plugin cannot be called from child module " +
          sonarProject.getGroupId() + ":" + sonarProject.getArtifactId() +
          ", please run it from the " + top.getGroupId() + ":" + top.getArtifactId() + " project POM");
      return;
    }
    firstCall = false;

    try {
      collect(sonarProject);
    } finally {
      manager.clear();
      if (!externalManager) {
        manager.close();
        manager = null;
      }
      runDate = null;
      webInterfaceService = null;
    }
  }

  private Persistence getPersistence() throws MojoExecutionException {
    Persistence persistence;
    try {
      persistence = Persistence.create(getJdbcData());
    } catch (PersistenceException e) {
      throw new MojoExecutionException("DB model exception occured", e);
    }
    return persistence;
  }

  private void collectAndPersist(Snapshot snapshot) throws MojoExecutionException {
    if (!mavenProject.getPackaging().equals("pom")) {
      if (filesRepository == null) {
        filesRepository = new FilesRepositoryImpl(getLog());
      }
      MetricService metricService = new MetricService(manager);
      RulesService rulesService = new RulesService(manager);
      for (Report report : targetReports) {
        MeasuresCollector collector = report.getMeasuresCollector();
        if (collector.initialize(mavenProject, metricService, rulesService, filesRepository)) {
          List<Collectable> collected = collector.collect();
          persistCollectables(snapshot, collected);
        }
      }
    }
  }

  private void collect(MavenProject sonarProject) throws MojoExecutionException {
    Snapshot snapshot = new Snapshot();
    snapshot.setCreatedAt(runDate);
    snapshot.setMavenProject(sonarProject);
    snapshot.setVersion(mavenProject.getVersion());
    manager.setFlushMode(FlushModeType.COMMIT);

    manager.getTransaction().begin();
    SnapshotGroup snapshotGroup;
    try {
      if (sonarProject.getParent() == null) {
        snapshotGroup = new SnapshotGroup();
        snapshotGroup.setCreatedAt(runDate);
        snapshotGroup.setMavenProject(sonarProject);
        manager.persist(snapshotGroup);
        setCurrentSnapshotGroupId(snapshotGroup.getId());
      } else {
        Integer groupId = getCurrentSnapshotGroupId();
        if (groupId != null) {
          snapshotGroup = manager.find(SnapshotGroup.class, groupId);
          if (snapshotGroup == null) {
            throw new NoResultException();
          }
        } else {
          SnapshotGroupService service = new SnapshotGroupService(manager);
          MavenProject parent = sonarProject.getParent();
          int maxDept = 100;
          while (maxDept > 0) {
            if (parent.getParent() == null) {
              break;
            }
            parent = parent.getParent();
            maxDept--;
          }
          // look for the last snapshot group created within the last 60 mins
          snapshotGroup = service.getLastUnprocessedGroup(parent.getId(), (long) 60 * 60 * 1000, runDate.getTime());
        }
      }

    } catch (NoResultException e) {
      throw new MojoExecutionException("Unable to find the last snapshot group");
    }
    snapshot.setSnapshotGroup(snapshotGroup);
    manager.persist(snapshot);
    collectAndPersist(snapshot);
    manager.getTransaction().commit();
    triggerMeasureCalculation(snapshotGroup);
  }

  private void triggerMeasureCalculation(SnapshotGroup snapshotGroup) throws MojoExecutionException {
    SnapshotGroupService snapshotGroupService = new SnapshotGroupService(manager);
    manager.refresh(snapshotGroup);
    if (snapshotGroupService.isReadyToCalculateMeasures(snapshotGroup, null)) {
      try {
        webInterfaceService.triggerMeasuresCalculations();
      } catch (IOException e) {
        throw new MojoExecutionException("Error occured when triggering measures calculation job on remote server", e);
      }
    }
  }

  private MavenProject saveMavenProject(org.apache.maven.project.MavenProject mavenProject, MavenProject parentProjet)
      throws XmlPullParserException, IOException {
    MavenProjectService service = new MavenProjectService(manager);
    MavenProject sonarProject;
    try {
      sonarProject = service.getMavenProject(mavenProject.getGroupId(), mavenProject.getArtifactId(), getBranch());
      updateMavenProject(sonarProject, mavenProject, parentProjet);
      manager.merge(sonarProject);
    } catch (NoResultException e) {
      sonarProject = new MavenProject();
      sonarProject.setArtifactId(mavenProject.getArtifactId());
      sonarProject.setGroupId(mavenProject.getGroupId());
      sonarProject.setBranch(getBranch());
      sonarProject.setEnabled(true);
      updateMavenProject(sonarProject, mavenProject, parentProjet);
      manager.persist(sonarProject);
    }
    saveModules(sonarProject, mavenProject);
    return sonarProject;

  }

  private void saveModules(MavenProject sonarProject, org.apache.maven.project.MavenProject mavenProject) throws IOException,
      XmlPullParserException {
    for (Object o : mavenProject.getModules()) {
      String moduleName = (String) o;
      MavenXpp3Reader pomReader = new MavenXpp3Reader();
      File moduleFile = new File(mavenProject.getBasedir() + "/" + moduleName + "/pom.xml");
      Model model = pomReader.read(new FileReader(moduleFile));
      org.apache.maven.project.MavenProject module = new org.apache.maven.project.MavenProject(model);
      if (module.getBasedir() == null) {
        module.setFile(moduleFile);
      }
      saveMavenProject(module, sonarProject);
    }
  }

  private void updateMavenProject(MavenProject sonarProject, org.apache.maven.project.MavenProject mavenProject, MavenProject parentProjet) {
    sonarProject.setName(mavenProject.getName());
    sonarProject.setDescription(mavenProject.getDescription());
    if (parentProjet != null) {
      sonarProject.setParent(parentProjet);
    }
    updateProjectLink(ProjectLink.LINK_HOME_PAGE, mavenProject.getUrl(), sonarProject);
    Scm scmConfig = mavenProject.getScm();
    if (scmConfig == null) {
      scmConfig = new Scm();
    }
    updateProjectLink(ProjectLink.LINK_SCM_URL, scmConfig.getUrl(), sonarProject);
    updateProjectLink(ProjectLink.LINK_SCM_DEV_CONNECTION, scmConfig.getDeveloperConnection(), sonarProject);
    updateProjectLink(ProjectLink.LINK_SCM_RO_CONNECTION, scmConfig.getConnection(), sonarProject);

    CiManagement cimConfig = mavenProject.getCiManagement();
    if (cimConfig == null) {
      cimConfig = new CiManagement();
    }
    updateProjectLink(ProjectLink.LINK_CONTINUOUS_INTEGRATION, cimConfig.getUrl(), sonarProject);

    IssueManagement imConfig = mavenProject.getIssueManagement();
    if (imConfig == null) {
      imConfig = new IssueManagement();
    }
    updateProjectLink(ProjectLink.LINK_ISSUES_TRACKER, imConfig.getUrl(), sonarProject);
  }

  private void updateProjectLink(String linkType, String href, MavenProject sonarProject) {
    if (href != null && !"".equals(href)) {
      ProjectLink link = sonarProject.getProjectLinkByType(linkType);
      if (link == null) {
        link = new ProjectLink();
        link.setMavenProject(sonarProject);
        link.setType(linkType);
        sonarProject.getProjectLinks().add(link);
      }
      link.setHref(href);
    }
  }
  
  private void persistCollectables(Snapshot snapshot, List<Collectable> collectables) {
    for (Collectable collectable : collectables) {
      if (collectable instanceof WithFile) {
        if (((WithFile) collectable).getFile() == null) {
          getLog().warn("FileMeasure without file : " + collectable + ", snapshot=" + snapshot.getId());
          continue;
        } else {
          ((WithFile) collectable).getFile().setSnapshot(snapshot);
        }
      }
      if (collectable instanceof MetricMeasure && ((MetricMeasure) collectable).getValue().isNaN()) {
        getLog().warn("Measure with NaN value : " + collectable + ", snapshot=" + snapshot.getId());
        continue;
      }
      collectable.setId(null);
      collectable.setSnapshot(snapshot);
      manager.persist(collectable);
    }
  }
}