/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with {library}; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package ch.hortis.sonar.mvn.mc;

import ch.hortis.sonar.model.Collectable;
import ch.hortis.sonar.model.File;
import ch.hortis.sonar.model.FileMeasure;
import ch.hortis.sonar.model.Metric;
import ch.hortis.sonar.model.Metrics;
import ch.hortis.sonar.model.ProjectMeasure;
import org.apache.commons.lang.StringUtils;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class JavaNCSSCollector extends BaseMeasuresCollector {
  protected XmlReportParser reportParser;
  protected Metric packageMetric;
  protected Metric ncssMetric;
  protected Metric complexityMetric;
  protected Metric classesMetric;
  protected Metric functionsMetric;

  public boolean initialize(MavenProject project) {
    java.io.File reportFile = findFileFromBuildDirectory(project, "javancss-raw-report.xml");
    if (reportFile != null) {
      reportParser = new XmlReportParser();
      reportParser.parse(reportFile);

      packageMetric = loadMetric(Metrics.NCSS_PACKAGES);
      ncssMetric = loadMetric(Metrics.NCSS_NCSS);
      complexityMetric = loadMetric(Metrics.CYCLOMATIC_COMPLEXITY);
      classesMetric = loadMetric(Metrics.NCSS_CLASSES);
      functionsMetric = loadMetric(Metrics.NCSS_FUNCTIONS);
      return true;
    }
    return false;
  }

  public List<Collectable> collect() throws MojoExecutionException {
    ArrayList<Collectable> result = new ArrayList<Collectable>();
    result.addAll(collectProjectMeasures());
    result.addAll(collectFileMeasures());
    return result;
  }

  public List<ProjectMeasure> collectProjectMeasures() throws MojoExecutionException {
    List<ProjectMeasure> measures = new ArrayList<ProjectMeasure>();
    try {
      measures.add(new ProjectMeasure(packageMetric, parseNumber(reportParser.executeXPath("count(/javancss/packages/package)"))));
      // warning bug : the total /javancss/packages/total/ncss does not match the sum of /javancss/objects/object/ncss so use the sum
      measures.add(new ProjectMeasure(ncssMetric, parseNumber(reportParser.executeXPath("sum(/javancss/objects/object/ncss/text())"))));
      measures.add(new ProjectMeasure(complexityMetric, parseNumber(reportParser.executeXPath("sum(/javancss/functions/function/ccn/text())"))));
      measures.add(new ProjectMeasure(classesMetric, parseNumber(reportParser.executeXPath("/javancss/packages/total/classes"))));
      measures.add(new ProjectMeasure(functionsMetric, parseNumber(reportParser.executeXPath("/javancss/packages/total/functions"))));

    } catch (ParseException ex) {
      throw new MojoExecutionException("ParseException during JavaNCSS report parsing", ex);
    }
    return measures;
  }

  public List<FileMeasure> collectFileMeasures() throws MojoExecutionException {
    try {
      ArrayList<FileMeasure> measures = new ArrayList<FileMeasure>();
      NodeList classes = reportParser.executeXPathNodeList("/javancss/objects/object");
      if (classes != null) {
        for (int i = 0; i < classes.getLength(); i++) {
          Element element = (Element) classes.item(i);
          String className = reportParser.getChildElementValue(element, "name");

          File file = getFileFromClassName(className);
          addFileMeasure(measures, file, element, "ncss", ncssMetric);
          addFileMeasure(measures, file, element, "functions", functionsMetric);
          addFileMeasure(measures, file, 1.0, classesMetric);
        }
      }
      ccnFileMeasures(measures);
      return measures;
    } catch (ParseException ex) {
      throw new MojoExecutionException("ParseException during JavaNCSS report parsing", ex);
    }
  }

  private void ccnFileMeasures(ArrayList<FileMeasure> measures)
      throws MojoExecutionException, ParseException {
    NodeList functions = reportParser.executeXPathNodeList("/javancss/functions/function");
    Map<String, Integer> ccnCountperClass = new HashMap<String, Integer>();
    if (functions != null) {
      for (int i = 0; i < functions.getLength(); i++) {
        Element element = (Element) functions.item(i);
        String functionName = reportParser.getChildElementValue(element, "name");
        Integer ccnForFunction = new Integer(reportParser.getChildElementValue(element, "ccn"));
        String className = getClassNameForFunction(functionName);
        Integer ccnSum = ccnCountperClass.get(className);
        if (ccnSum == null) {
          ccnSum = 0;
        }
        ccnSum += ccnForFunction;
        ccnCountperClass.put(className, ccnSum);
      }
    }
    for (String className : ccnCountperClass.keySet()) {
      Integer ccnForClass = ccnCountperClass.get(className);
      File file = getFileFromClassName(className);
      addFileMeasure(measures, file, ccnForClass.doubleValue(), complexityMetric);
    }
  }
  
  protected String getClassNameForFunction(String functionName) {
    String className = StringUtils.substringBeforeLast(functionName, ".");
    return getClassName( className );
  }
  
  private String getClassName(String classNameCandidate) {

    int lastPointIndex = classNameCandidate.lastIndexOf( "." );
    if ( lastPointIndex == -1 ) {
      return classNameCandidate;
    }
    String packageName = StringUtils.substringBeforeLast(classNameCandidate, ".");
    String className = StringUtils.substringAfterLast(classNameCandidate, ".");
    String[] packages = StringUtils.split( packageName, '.' );
    for (String pack : packages) {
      if ( !isPackageName(pack) ) {
        return getClassName(packageName);
      }
    }
    return packageName + "." + className;
  }
  
  protected boolean isPackageName(String classNameCandidate) {
    if (!Character.isLowerCase(classNameCandidate.charAt(0))) {
      return false;
    }
    for (int i = 0; i < classNameCandidate.length(); i++) {
      char currentLetter = classNameCandidate.charAt(i);
      if (!Character.isLetterOrDigit(currentLetter)) {
        return false;
      }
    }
    return true;
  }

  private void addFileMeasure(ArrayList<FileMeasure> measures, File file,
                              Element classElement, String key, Metric metric) throws ParseException {
    Element element = reportParser.getChildElement(classElement, key);
    if (element != null && element.getFirstChild() != null && element.getFirstChild().getNodeValue() != null) {
      String value = element.getFirstChild().getNodeValue();
      addFileMeasure(measures, file, parseNumber(value), metric);
    }
  }

  private void addFileMeasure(ArrayList<FileMeasure> measures, File file, Double value, Metric metric) throws ParseException {
    FileMeasure fileMeasure = new FileMeasure();
    fileMeasure.setFile(file);
    fileMeasure.setValue(value);
    fileMeasure.setMetric(metric);
    measures.add(fileMeasure);
  }
}
